#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>
#include <float.h>

// #define EXTRAPRECISION

#ifdef EXTRAPRECISION
	typedef long double REAL;
	#define MPIREAL MPI_LONG_DOUBLE
#else
	typedef double REAL;
	#define MPIREAL MPI_DOUBLE
#endif


REAL area(long id, long nb)
{
	REAL x = (REAL) id / nb;
	return 4.0 / (1.0+x*x) / nb;
}


int main(int argc, char* argv[])
{
	long samples = 1e8;
	if (argc > 1) samples = atol(argv[1]);

	// Context initialization
	MPI_Init(&argc, &argv);
	int rank, size;
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &size);

	double time_start = MPI_Wtime();

	// Per process computation
	REAL subtotal;
	for (long i=rank; i<samples; i+=size) subtotal += area(i, samples);

	// Result reduction
	REAL total;
	MPI_Reduce(&subtotal, &total, 1, MPIREAL, MPI_SUM, 0, MPI_COMM_WORLD);

	// Evaluating performances
	double time_spent = MPI_Wtime() - time_start;
	double time_min, time_sum, time_max;
	MPI_Reduce(&time_spent, &time_min, 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD);
	MPI_Reduce(&time_spent, &time_sum, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
	MPI_Reduce(&time_spent, &time_max, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);

	// Printing results
	if (!rank)
	{
		#ifdef EXTRAPRECISION
			fprintf(stderr, "Pi = %17.15Lf\n", total);
		#else
			fprintf(stderr, "Pi = %17.15f\n",  total);
		#endif
		fprintf(stdout, "%ld %d --- %f %f %f --- %f\n", samples, size, time_min, time_sum/size, time_max, time_sum);
	}


	MPI_Finalize();

	return 0;
}
